/**************************************
/ SFP-Ethernet Example Design 
/ Rice WARP Project
/ (c) Copyright 2009 Rice University
/ See http://warp.rice.edu/license
/**************************************/

#include "xparameters.h"
#include "xlltemac.h"
#include "xllfifo.h"
#include "xexception_l.h"

#define TEMAC0_DEVICE_ID   XPAR_LLTEMAC_0_DEVICE_ID
#define FIFO0_DEVICE_ID    XPAR_LLFIFO_0_DEVICE_ID

#define TEMAC1_DEVICE_ID   XPAR_LLTEMAC_1_DEVICE_ID
#define FIFO1_DEVICE_ID    XPAR_LLFIFO_1_DEVICE_ID

#define TEST_FRAME_LEN 1400
#define NUM_TEST_FRAMES 1000

//Declare instances of the TEMAC driver's config struct
// Each TEMAC needs one
XLlTemac TemacInstance0;
XLlTemac TemacInstance1;

//Declare instances of the LL_FIFO driver's config struct
// Each FIFO needs one
XLlFifo FifoInstance0;
XLlFifo FifoInstance1;

//Function declarations
void TxRxTest(void);
int RxPkt(XLlFifo* fifoPtr, XLlTemac* temacPtr, void* pkt);
void TxPkt(XLlFifo* fifoPtr, void* pkt, int length);

int main()
{
	int Status;
	XLlTemac_Config* MacCfgPtr0;
	XLlTemac_Config* MacCfgPtr1;
	
	xil_printf("SFP-EMAC Loopback Design v0 - Entering Main\r\n");

	//Get pointers for each TEMAC configuration
	MacCfgPtr0 = XLlTemac_LookupConfig(TEMAC0_DEVICE_ID);
	MacCfgPtr1 = XLlTemac_LookupConfig(TEMAC1_DEVICE_ID);
	
	//Initialize the TEMAC configurations
	Status = XLlTemac_CfgInitialize(&TemacInstance0, MacCfgPtr0, MacCfgPtr0->BaseAddress);
	Status |= XLlTemac_CfgInitialize(&TemacInstance1, MacCfgPtr1, MacCfgPtr1->BaseAddress);

	if (Status != XST_SUCCESS)
	{
			xil_printf("Error in initialize\r\n");
			return XST_FAILURE;
	}

	//Poll the TEMAC status bits indicating whether the MGT cores have locked to their reference clock
	xil_printf("Waiting for MGTs to lock\r\n");
	Status = 0;
	while(Status == 0)
	{
		Status = (XLlTemac_ReadReg(TemacInstance0.Config.BaseAddress, XTE_IS_OFFSET) | XLlTemac_ReadReg(TemacInstance1.Config.BaseAddress, XTE_IS_OFFSET)) & XTE_INT_MGT_LOCK_MASK;
	}

	//Set some sensible options for the TEMAC
	//Some of these are supposed to be set by default in the driver, but it can't hurt
	// to re-set them here
   	Status  = XLlTemac_ClearOptions(&TemacInstance0, XTE_LENTYPE_ERR_OPTION | XTE_FLOW_CONTROL_OPTION);
	Status |= XLlTemac_SetOptions(&TemacInstance0, XTE_PROMISC_OPTION | XTE_MULTICAST_OPTION | XTE_BROADCAST_OPTION | XTE_RECEIVER_ENABLE_OPTION | XTE_TRANSMITTER_ENABLE_OPTION);
   	Status |= XLlTemac_ClearOptions(&TemacInstance1, XTE_LENTYPE_ERR_OPTION | XTE_FLOW_CONTROL_OPTION);
	Status |= XLlTemac_SetOptions(&TemacInstance1, XTE_PROMISC_OPTION | XTE_MULTICAST_OPTION | XTE_BROADCAST_OPTION | XTE_RECEIVER_ENABLE_OPTION | XTE_TRANSMITTER_ENABLE_OPTION);
	if (Status != XST_SUCCESS)
	{
		xil_printf("Error setting EMAC options\r\n, code %d", Status);
		return XST_FAILURE;
	}

	//Initialize the LL_FIFO cores
	XLlFifo_Initialize(&FifoInstance0, XLlTemac_LlDevBaseAddress(&TemacInstance0));
	XLlFifo_Initialize(&FifoInstance1, XLlTemac_LlDevBaseAddress(&TemacInstance1));

	//Make sure both TEMACs are configured for FIFO mode (and not MPMC/DMA)
	if (!XLlTemac_IsFifo(&TemacInstance0) | !XLlTemac_IsFifo(&TemacInstance1))
	{
			xil_printf("Device HW not configured for FIFO direct mode\r\n");
			return XST_FAILURE;
	}

	xil_printf("Waiting for TEMACs to be ready...");

	do
	{
		Status = XLlTemac_ReadReg(TemacInstance0.Config.BaseAddress, XTE_RDY_OFFSET);
	} while ((Status & XTE_RDY_HARD_ACS_RDY_MASK) == 0);

	do
	{
		Status = XLlTemac_ReadReg(TemacInstance1.Config.BaseAddress, XTE_RDY_OFFSET);
	} while ((Status & XTE_RDY_HARD_ACS_RDY_MASK) == 0);

	xil_printf("done\r\n");

	XLlTemac_SetOperatingSpeed(&TemacInstance0, 1000);
	XLlTemac_SetOperatingSpeed(&TemacInstance1, 1000);
	sleep(1);

	xil_printf("Starting TEMACs...");
    XLlTemac_Start(&TemacInstance0);
    XLlTemac_Start(&TemacInstance1);

	do
	{
		Status = XLlTemac_IsStarted(&TemacInstance0);
	} while (Status == 0);

	do
	{
		Status = XLlTemac_IsStarted(&TemacInstance1);
	} while (Status == 0);
	xil_printf("done\r\n");
	

	xil_printf("Waiting for PHY to start...");
	sleep(5);
	xil_printf("done.\r\n");

	xil_printf("Starting SFP Ethernet Loopback Test\r\n");

	TxRxTest();

	xil_printf("Done. Exiting...\r\n");

	return 0;
}

void TxPkt(XLlFifo* fifoPtr, void* pkt, int length)
{
	XLlFifo_Write(fifoPtr, pkt, length);
	XLlFifo_TxSetLen(fifoPtr, length);
	
	while(  (XLlFifo_Status(fifoPtr) & XLLF_INT_TC_MASK) == 0) {}
	XLlFifo_IntClear(fifoPtr, XLLF_INT_TC_MASK);

	return;
}

int RxPkt(XLlFifo* fifoPtr, XLlTemac* temacPtr, void* pkt)
{
	int Status;
	int RxFrameLength;

	Status = XST_NO_DATA;
	do
	{
		if (XLlFifo_Status(fifoPtr) & XLLF_INT_RC_MASK) {
			XLlFifo_IntClear(fifoPtr, XLLF_INT_RC_MASK);
			Status = XST_SUCCESS;
		}
		if (XLlFifo_Status(fifoPtr) & XLLF_INT_ERROR_MASK) {
			XLlFifo_IntClear(fifoPtr, XLLF_INT_ERROR_MASK);
			Status = XST_FIFO_ERROR;
			xil_printf("FIFO error\r\n");
			break;
		}
		if (XLlTemac_Status(temacPtr) & XTE_INT_RXRJECT_MASK) {
			xil_printf("Dropped frame! 3\r\n");
			XLlTemac_IntClear(temacPtr, (XTE_INT_RXFIFOOVR_MASK | XTE_INT_RXRJECT_MASK));
		}
		/* When the RXFIFOOVR bit is set, the RXRJECT bit also gets set */
		if (XLlTemac_Status(temacPtr) & XTE_INT_RXFIFOOVR_MASK) {
			xil_printf("Dropped frame! 4\r\n");
			XLlTemac_IntClear(temacPtr, (XTE_INT_RXFIFOOVR_MASK ));
		}
	} while (Status == XST_NO_DATA);


	if (Status == XST_SUCCESS)
	{
		RxFrameLength = XLlFifo_RxGetLen(fifoPtr);
		
		/* Read it out */
		XLlFifo_Read(fifoPtr, pkt, RxFrameLength);
		
		if (XLlTemac_Status(temacPtr) & XTE_INT_RXRJECT_MASK) {
			xil_printf("Dropped frame! 1\r\n");
			XLlTemac_IntClear(temacPtr, (XTE_INT_RXFIFOOVR_MASK | XTE_INT_RXRJECT_MASK));
			return -1;
		}
		else
		{
			return RxFrameLength;
		}
	}		
	else
	{
		xil_printf("ERROR: Fatal error detected, code %d\r\n", XLlTemac_Status(temacPtr));
		return -1;
	}
}

void TxRxTest(void)
{
	int i, n;
	int Status;
	int RxFrameLen;
	u32 FifoFreeBytes;
	int pktTxNum = 0;
	
	unsigned char TxFrame[TEST_FRAME_LEN] __attribute__ ((aligned(8)));
	unsigned char RxFrame[TEST_FRAME_LEN] __attribute__ ((aligned(8)));

	XLlFifo* txFifo;
	XLlFifo* rxFifo;
	XLlTemac* rxTemac;

	for(i=0; i<14; i++)
	{
		TxFrame[i] = 0xFF;
	}

	xil_printf("\r\n\r\nMGT #7 -> MGT #8 Frame Test: 000");

	txFifo = &FifoInstance0;
	rxFifo = &FifoInstance1;
	rxTemac = &TemacInstance1;

	for(n=0; n<NUM_TEST_FRAMES; n++)
	{
		xil_printf("\b\b\b%03d", n);
	
		//Create a random payload
		for(i=14; i<TEST_FRAME_LEN; i++)
		{
			TxFrame[i] = 0xFF & rand();
		}
	
		TxPkt(txFifo, TxFrame, TEST_FRAME_LEN);
		RxFrameLen = RxPkt(rxFifo, rxTemac, RxFrame);
	
		if(RxFrameLen == TEST_FRAME_LEN)
		{
			Status = 0;
			for(i=0; i<TEST_FRAME_LEN; i++)
				Status += (RxFrame[i] - TxFrame[i]);
			
			if(Status)
			{
				xil_printf("\r\nERROR: Tx and Rx frames don't match!\r\n");
				return;
			}
		}
		else
		{
			xil_printf("\r\nERROR: Tx and Rx frames lengths don't match!\r\n");
			return;
		}
	}

	xil_printf("\r\nMGT #7 -> MGT #8 Tested Successfully\r\n\r\n");

	xil_printf("\r\nMGT #8 -> MGT #7 Frame Test: 000");

	txFifo = &FifoInstance1;
	rxFifo = &FifoInstance0;
	rxTemac = &TemacInstance0;

	for(n=0; n<NUM_TEST_FRAMES; n++)
	{
		xil_printf("\b\b\b%03d", n);
	
		//Create a random payload
		for(i=14; i<TEST_FRAME_LEN; i++)
		{
			TxFrame[i] = 0xFF & rand();
		}
	
		TxPkt(txFifo, TxFrame, TEST_FRAME_LEN);
		RxFrameLen = RxPkt(rxFifo, rxTemac, RxFrame);
	
		if(RxFrameLen == TEST_FRAME_LEN)
		{
			Status = 0;
			for(i=0; i<TEST_FRAME_LEN; i++)
				Status += (RxFrame[i] - TxFrame[i]);
			
			if(Status)
			{
				xil_printf("\r\nERROR: Tx and Rx frames don't match!\r\n");
				return;
			}
		}
		else
		{
			xil_printf("\r\nERROR: Tx and Rx frames lengths don't match!\r\n");
			return;
		}
	}

	xil_printf("\r\nMGT #8 -> MGT #7 Tested Successfully\r\n\r\n");
}
